/*
 *  RSGISApplyGainOffset2Img.h
 *  RSGIS_LIB
 *
 *  Created by Pete Bunting on 16/03/2017.
 *  Copyright 2017 RSGISLib. All rights reserved.
 *  This file is part of RSGISLib.
 *
 *  RSGISLib is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  RSGISLib is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef RSGISApplyGainOffset2Img_H
#define RSGISApplyGainOffset2Img_H

#include <iostream>
#include <stdio.h>
#include <cmath>

#include "img/RSGISCalcImage.h"
#include "img/RSGISCalcImageValue.h"
#include "img/RSGISImageBandException.h"
#include "img/RSGISImageCalcException.h"



// mark all exported classes/functions with DllExport to have
// them exported by Visual Studio
#undef DllExport
#ifdef _MSC_VER
#ifdef rsgis_img_EXPORTS
#define DllExport   __declspec( dllexport )
#else
#define DllExport   __declspec( dllimport )
#endif
#else
#define DllExport
#endif

namespace rsgis{namespace img{
    
    
    class DllExport RSGISRescaleImageData : public RSGISCalcImageValue
    {
    public:
        RSGISRescaleImageData(int numOutputBands, float cNoDataVal, float cOffset, float cGain, float nNoDataVal, float nOffset, float nGain);
        void calcImageValue(float *bandValues, int numBands, double *output);
        ~RSGISRescaleImageData();
    protected:
        float cNoDataVal;
        float cOffset;
        float cGain;
        float nNoDataVal;
        float nOffset;
        float nGain;
    };
    
}}

#endif
