/*
 *  RSGISEliminateSinglePixels.h
 *  RSGIS_LIB
 *
 *  Created by Pete Bunting on 03/05/2012.
 *  Copyright 2012 RSGISLib.
 * 
 *  RSGISLib is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  RSGISLib is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef RSGISEliminateSinglePixels_H
#define RSGISEliminateSinglePixels_H

#include <iostream>
#include <string>
#include <cmath>
#include <stdlib.h>

#include "common/rsgis-tqdm.h"

#include "img/RSGISImageUtils.h"
#include "img/RSGISImageCalcException.h"
#include "math/RSGISMathsUtils.h"

#include "img/RSGISCalcImageValue.h"
#include "img/RSGISCalcImage.h"

#include "rastergis/RSGISRasterAttUtils.h"

#include "gdal_priv.h"
#include "ogrsf_frmts.h"
#include "ogr_api.h"

// mark all exported classes/functions with DllExport to have
// them exported by Visual Studio
#undef DllExport
#ifdef _MSC_VER
    #ifdef rsgis_segmentation_EXPORTS
        #define DllExport   __declspec( dllexport )
    #else
        #define DllExport   __declspec( dllimport )
    #endif
#else
    #define DllExport
#endif

namespace rsgis{namespace segment{
    
    class DllExport RSGISEliminateSinglePixels
    {
    public:
        RSGISEliminateSinglePixels();
        void eliminate(GDALDataset *inSpecData, GDALDataset *inClumpsData, GDALDataset *tmpData, std::string outputImage, float noDataVal, bool noDataValProvided, bool projFromImage, std::string proj, std::string format);
        void eliminateBlocks(GDALDataset *inSpecData, GDALDataset *inClumpsData, GDALDataset *tmpData, std::string outputImage, float noDataVal, bool noDataValProvided, bool projFromImage, std::string proj, std::string format);
        ~RSGISEliminateSinglePixels();
    private:
        unsigned long findSinglePixels(GDALDataset *inClumpsData, GDALDataset *tmpData, float noDataVal, bool noDataValProvided);
        bool eliminateSinglePixels(GDALDataset *inSpecData, GDALDataset *inClumpsData, GDALDataset *tmpData, GDALDataset *outDataset, float noDataVal, bool noDataValProvided);
        inline float eucDistance(float **vals1, float **vals2, unsigned int numBands, unsigned int col1, unsigned int col2);
    };
    
    
    
    class DllExport RSGISFindSinglePixels : public rsgis::img::RSGISCalcImageValue
	{
	public:
		RSGISFindSinglePixels(float noDataVal, bool noDataValProvided);
		void calcImageValue(float ***dataBlock, int numBands, int winSize, double *output);
		void resetCount(){this->numSingles = 0;};
        unsigned long getCount(){return this->numSingles;};
        ~RSGISFindSinglePixels();
    protected:
        float noDataVal;
        bool noDataValProvided;
        unsigned long numSingles;
	};
    
    
    class DllExport RSGISElimSinglePixelsCalcImg : public rsgis::img::RSGISCalcImageValue
	{
	public:
		RSGISElimSinglePixelsCalcImg(float noDataVal, bool noDataValProvided);
		void calcImageValue(float ***dataBlock, int numBands, int winSize, double *output);
		void resetChangeOccured(){this->hasChangeOccured = false;};
        bool getChangeOccured(){return this->hasChangeOccured;};
        ~RSGISElimSinglePixelsCalcImg();
    protected:
        bool hasChangeOccured;
        float noDataVal;
        bool noDataValProvided;
        inline float eucDistance(float *vals1, float *vals2, unsigned int numBands);
	};
    
}}

#endif
